/*
 * Interprocess communication abstractions.
 *
 * Copyright (C) 2018-2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/util/util.h>
#include <l4/sys/factory.h>
#include <l4/sys/irq.h>

#include "cap_alloc.h"
#include "mem_ipc.h"
#include "util_ipc.h"



/* Declare expected capabilities.
   NOTE: This does not support the mixing of expected capabilities and other
         items. */

long _expect_capabilities(l4_buf_regs_t *bregs, int number)
{
  int i;
  long err;

  for (i = 0; i < number; i++)
  {
    err = _expect_capability(bregs, i);
    if (err)
      return err;
  }

  return L4_EOK;
}

/* Indicate that a capability is expected at the given position. */

long _expect_capability(l4_buf_regs_t *bregs, int item)
{
  l4_cap_idx_t future = ipc_cap_alloc();

  if (l4_is_invalid_cap(future))
    return -L4_ENOENT;

  /* Indicate the expectation of a capability in return. */

  bregs->bdr = 0;
  bregs->br[item] = L4_RCV_ITEM_SINGLE_CAP | future | L4_RCV_ITEM_LOCAL_ID;

  return L4_EOK;
}

/* Indicate that a flexpage is expected at the given position. */

long _expect_fpage(l4_buf_regs_t *bregs, int item, l4_umword_t map_control, l4_fpage_t fpage)
{
  bregs->bdr = 0;
  bregs->br[item] = map_control;
  bregs->br[item + 1] = fpage.raw;

  return L4_EOK;
}

/* Export in the message at the given position the given capability. */

void _export_capability(l4_msgtag_t tag, l4_msg_regs_t *mregs, int item, l4_cap_idx_t ref)
{
  int words = l4_msgtag_words(tag);

  mregs->mr[words + item * 2] = 0 | L4_ITEM_MAP;
  mregs->mr[words + item * 2 + 1] = l4_obj_fpage(ref, 0, L4_FPAGE_RWX).raw;
}

/* Export a flexpage at the given position in the message. Here, the snd_base
   member of the flexpage structure is used to hold the "hot spot" value. */

void _export_fpage(l4_msgtag_t tag, l4_msg_regs_t *mregs, int item, l4_snd_fpage_t fpage)
{
  int words = l4_msgtag_words(tag);

  mregs->mr[words + item * 2] = l4_map_control(fpage.snd_base, L4_FPAGE_CACHEABLE, 0);
  mregs->mr[words + item * 2 + 1] = fpage.fpage.raw;
}

/* Export a flexpage at the given position in the message. */

void _export_page(l4_msgtag_t tag, l4_msg_regs_t *mregs, int item, l4_umword_t hot_spot, l4_fpage_t fpage)
{
  int words = l4_msgtag_words(tag);

  mregs->mr[words + item * 2] = l4_map_control(hot_spot, L4_FPAGE_CACHEABLE, 0);
  mregs->mr[words + item * 2 + 1] = fpage.raw;
}

/* Free expected capabilities held by the buffer registers. */

void _free_expected_capabilities(l4_buf_regs_t *bregs, int start, int limit)
{
  int i;
  l4_cap_idx_t cap;

  for (i = start; i < limit; i++)
  {
    cap = bregs->br[i] & L4_CAP_MASK;
    if (l4_is_valid_cap(cap))
      ipc_cap_free_um(cap);
  }
}

/* Import from the message the capability at the given item position. */

long _import_capability(l4_msgtag_t tag, l4_buf_regs_t *bregs,
                        l4_msg_regs_t *mregs, int item, int buf_item,
                        l4_cap_idx_t *ref, int *local)
{
  l4_snd_fpage_t fpage;
  long err = _import_fpage(tag, bregs, mregs, item, &fpage);

  if (err)
    return err;

  /* Check for a received capability and update the supplied capability
     index. */

  /* Inter-task capability transfer. */

  if ((fpage.snd_base & 0x3e) == 0x38)
  {
    *ref = bregs->br[buf_item] & L4_CAP_MASK;
    *local = 0;
  }

  /* Local capability transfer. */

  else if ((fpage.snd_base & 0x3e) == 0x3e)
  {
    *ref = l4_fpage_obj(fpage.fpage);
    *local = 1;
  }

  /* Unsupported item which may legitimately be an invalid capability. */

  else
    *ref = L4_INVALID_CAP;

  return L4_EOK;
}

/* Import from the message a dataspace, mapping it to an address. */

long _import_dataspace(l4_msgtag_t tag, l4_buf_regs_t *bregs,
                       l4_msg_regs_t *mregs, int item, int buf_item,
                       l4re_ds_t *mem, l4_addr_t *addr, int *local)
{
  long err;
  unsigned long size;

  *mem = ipc_cap_alloc();

  err = _import_capability(tag, bregs, mregs, item, buf_item, mem, local);

  if (err)
  {
    if (!*local)
      ipc_cap_free_um(*mem);
    return err;
  }

  /* Attach the dataspace to a region. */

  err = ipc_dataspace_size(*mem, &size);
  if (err)
  {
    ipc_cap_free_um(*mem);
    return err;
  }

  err = ipc_attach_dataspace(*mem, size, (void **) addr);
  if (err)
  {
    ipc_cap_free_um(*mem);
    return err;
  }

  return L4_EOK;
}

/* Import from the message the flexpage at the given item position. */

long _import_fpage(l4_msgtag_t tag, l4_buf_regs_t *bregs,
                   l4_msg_regs_t *mregs, int item, l4_snd_fpage_t *fpage)
{
  int words = l4_msgtag_words(tag);

  /* NOTE: Currently, this function does not interact with the received items,
           but instead just returns the flexpage details from the message
           words. */

  (void) bregs;

  /* Obtain the item details for the capability. */

  if (item >= (int) l4_msgtag_items(tag))
    return -L4_EIO;

  /* Received items should start at zero. */

  fpage->snd_base = mregs->mr[words + item * 2];
  fpage->fpage.raw = mregs->mr[words + item * 2 + 1];

  return L4_EOK;
}



/* Convenience operations. */

/* Send a reply to a call using the given message tag. */

l4_msgtag_t util_ipc_reply(l4_msgtag_t tag)
{
  return l4_ipc_send(L4_INVALID_CAP | L4_SYSF_REPLY, l4_utcb(), tag, L4_IPC_NEVER);
}

/* Send a request using a call with the given message tag to the indicated
   endpoint. */

l4_msgtag_t util_ipc_request(l4_msgtag_t tag, l4_cap_idx_t endpoint)
{
  return l4_ipc_call(endpoint, l4_utcb(), tag, L4_IPC_NEVER);
}

/* Send data with the given message tag to the indicated endpoint. */

l4_msgtag_t util_ipc_send(l4_msgtag_t tag, l4_cap_idx_t endpoint)
{
  return l4_ipc_send(endpoint, l4_utcb(), tag, L4_IPC_NEVER);
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
